@@ -3,7 +3,6 @@
 {-# LANGUAGE EmptyCase #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE NamedFieldPuns #-}
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE TemplateHaskellQuotes #-}
 
@@ -111,45 +110,30 @@ module Data.Aeson.TH
     , mkLiftParseJSON2
     ) where
 
-import Prelude.Compat hiding (fail)
-
 -- We don't have MonadFail Q, so we should use `fail` from real `Prelude`
-import Prelude (fail)
 
-import Control.Applicative ((<|>))
+import Data.Aeson.Internal.Prelude
+
 import Data.Char (ord)
 import Data.Aeson (Object, (.:), FromJSON(..), FromJSON1(..), FromJSON2(..), ToJSON(..), ToJSON1(..), ToJSON2(..))
 import Data.Aeson.Types (Options(..), Parser, SumEncoding(..), Value(..), defaultOptions, defaultTaggedObject)
 import Data.Aeson.Types.Internal ((<?>), JSONPathElement(Key))
-import Data.Aeson.Types.FromJSON (parseOptionalFieldWith)
 import Data.Aeson.Types.ToJSON (fromPairs, pair)
 import Data.Aeson.Key (Key)
 import qualified Data.Aeson.Key as Key
 import qualified Data.Aeson.KeyMap as KM
-import Control.Monad (liftM2, unless, when)
 import Data.Foldable (foldr')
-#if MIN_VERSION_template_haskell(2,8,0) && !MIN_VERSION_template_haskell(2,10,0)
-import Data.List (nub)
-#endif
-import Data.List (foldl', genericLength, intercalate, partition, union)
+import Data.List (genericLength, intercalate, union)
 import Data.List.NonEmpty ((<|), NonEmpty((:|)))
 import Data.Map (Map)
-import Data.Maybe (catMaybes, fromMaybe, mapMaybe)
 import qualified Data.Monoid as Monoid
 import Data.Set (Set)
 import Language.Haskell.TH hiding (Arity)
-import Language.Haskell.TH.Datatype as Datatype
-#if MIN_VERSION_template_haskell(2,8,0) && !(MIN_VERSION_template_haskell(2,10,0))
-import Language.Haskell.TH.Syntax (mkNameG_tc)
-#endif
+import Language.Haskell.TH.Datatype
 import Text.Printf (printf)
 import qualified Data.Aeson.Encoding.Internal as E
-import qualified Data.Foldable as F (all)
 import qualified Data.List.NonEmpty as NE (length, reverse)
 import qualified Data.Map as M (fromList, keys, lookup , singleton, size)
-#if !MIN_VERSION_base(4,16,0)
-import qualified Data.Semigroup as Semigroup (Option(..))
-#endif
 import qualified Data.Set as Set (empty, insert, member)
 import qualified Data.Text as T (pack, unpack)
 import qualified Data.Vector as V (unsafeIndex, null, length, create, empty)
@@ -336,10 +320,11 @@ consToValue _ _ _ _ [] =
 
 consToValue target jc opts instTys cons = autoletE liftSBS $ \letInsert -> do
     value <- newName "value"
+    os    <- newNameList "_o"   $ arityInt jc
     tjs   <- newNameList "_tj"  $ arityInt jc
     tjls  <- newNameList "_tjl" $ arityInt jc
-    let zippedTJs      = zip tjs tjls
-        interleavedTJs = interleave tjs tjls
+    let zippedTJs      = zip3 os tjs tjls
+        interleavedTJs = flatten3 zippedTJs
         lastTyVars     = map varTToName $ drop (length instTys - arityInt jc) instTys
         tvMap          = M.fromList $ zip lastTyVars zippedTJs
     lamE (map varP $ interleavedTJs ++ [value]) $
@@ -468,40 +453,26 @@ argsToValue letInsert target jc tvMap opts multiCons
       (True,True,[_]) -> argsToValue letInsert target jc tvMap opts multiCons
                                      (info{constructorVariant = NormalConstructor})
       _ -> do
+
         argTys' <- mapM resolveTypeSynonyms argTys
         args <- newNameList "arg" $ length argTys'
-        let pairs | omitNothingFields opts = infixApp maybeFields
-                                                      [|(Monoid.<>)|]
-                                                      restFields
-                  | otherwise = mconcatE (map pureToPair argCons)
+        let argCons = zip3 (map varE args) argTys' fields
 
-            argCons = zip3 (map varE args) argTys' fields
+            toPair (arg, argTy, fld) =
+              let fieldName = fieldLabel opts fld
+                  toValue = dispatchToJSON target jc conName tvMap argTy
 
-            maybeFields = mconcatE (map maybeToPair maybes)
+                  omitFn :: Q Exp
+                  omitFn
+                    | omitNothingFields opts = dispatchOmitField jc conName tvMap argTy
+                    | otherwise = [| const False |]
 
-            restFields = mconcatE (map pureToPair rest)
+              in condE
+                (omitFn `appE` arg)
+                [| mempty |]
+                (pairE letInsert target fieldName (toValue `appE` arg))
 
-            (maybes0, rest0) = partition isMaybe argCons
-#if MIN_VERSION_base(4,16,0)
-            maybes = maybes0
-            rest   = rest0
-#else
-            (options, rest) = partition isOption rest0
-            maybes = maybes0 ++ map optionToMaybe options
-#endif
-
-            maybeToPair = toPairLifted True
-            pureToPair = toPairLifted False
-
-            toPairLifted lifted (arg, argTy, field) =
-              let toValue = dispatchToJSON target jc conName tvMap argTy
-                  fieldName = fieldLabel opts field
-                  e arg' = pairE letInsert target fieldName (toValue `appE` arg')
-              in if lifted
-                then do
-                  x <- newName "x"
-                  [|maybe mempty|] `appE` lam1E (varP x) (e (varE x)) `appE` arg
-                else e arg
+            pairs = mconcatE (map toPair argCons)
 
         match (conP conName $ map varP args)
               (normalB $ recordSumToValue letInsert target opts multiCons (null argTys) conName pairs)
@@ -526,19 +497,6 @@ argsToValue letInsert target jc tvMap opts multiCons
           )
           []
 
-isMaybe :: (a, Type, b) -> Bool
-isMaybe (_, AppT (ConT t) _, _) = t == ''Maybe
-isMaybe _                       = False
-
-#if !MIN_VERSION_base(4,16,0)
-isOption :: (a, Type, b) -> Bool
-isOption (_, AppT (ConT t) _, _) = t == ''Semigroup.Option
-isOption _                       = False
-
-optionToMaybe :: (ExpQ, b, c) -> (ExpQ, b, c)
-optionToMaybe (a, b, c) = ([|Semigroup.getOption|] `appE` a, b, c)
-#endif
-
 (<^>) :: ExpQ -> ExpQ -> ExpQ
 (<^>) a b = infixApp a [|(E.><)|] b
 infixr 6 <^>
@@ -697,10 +655,11 @@ consFromJSON _ _ _ _ [] =
 
 consFromJSON jc tName opts instTys cons = do
   value <- newName "value"
+  os    <- newNameList "_o"   $ arityInt jc
   pjs   <- newNameList "_pj"  $ arityInt jc
   pjls  <- newNameList "_pjl" $ arityInt jc
-  let zippedPJs      = zip pjs pjls
-      interleavedPJs = interleave pjs pjls
+  let zippedPJs      = zip3 os pjs pjls
+      interleavedPJs = flatten3 zippedPJs
       lastTyVars     = map varTToName $ drop (length instTys - arityInt jc) instTys
       tvMap          = M.fromList $ zip lastTyVars zippedPJs
   lamE (map varP $ interleavedPJs ++ [value]) $ lamExpr value tvMap
@@ -965,6 +924,11 @@ parseRecord jc tvMap argTys opts tName conName fields obj inTaggedObject =
            (infixApp (conE conName) [|(<$>)|] x)
            xs
     where
+      lookupField :: Type -> Q Exp
+      lookupField argTy
+        | allowOmittedFields opts = [| lookupFieldOmit |] `appE` dispatchOmittedField jc conName tvMap argTy
+        | otherwise               = [| lookupFieldNoOmit |]
+
       tagFieldNameAppender =
           if inTaggedObject then (tagFieldName (sumEncoding opts) :) else id
       knownFields = appE [|KM.fromList|] $ listE $
@@ -981,7 +945,7 @@ parseRecord jc tvMap argTys opts tName conName fields obj inTaggedObject =
                           (appE [|show|] (varE unknownFields)))
                       []
               ]
-      x:xs = [ [|lookupField|]
+      x:xs = [ lookupField argTy
                `appE` dispatchParseJSON jc conName tvMap argTy
                `appE` litE (stringL $ show tName)
                `appE` litE (stringL $ constructorTagModifier opts $ nameBase conName)
@@ -1149,26 +1113,17 @@ parseTypeMismatch tName conName expected actual =
           , actual
           ]
 
-class LookupField a where
-    lookupField :: (Value -> Parser a) -> String -> String
-                -> Object -> Key -> Parser a
+lookupFieldOmit :: Maybe a -> (Value -> Parser a) -> String -> String -> Object -> Key -> Parser a
+lookupFieldOmit maybeDefault pj tName rec obj key =
+    case KM.lookup key obj of
+      Nothing ->
+        case maybeDefault of
+          Nothing -> unknownFieldFail tName rec (Key.toString key)
+          Just x -> pure x
+      Just v  -> pj v <?> Key key
 
-instance {-# OVERLAPPABLE #-} LookupField a where
-    lookupField = lookupFieldWith
-
-instance {-# INCOHERENT #-} LookupField (Maybe a) where
-    lookupField pj _ _ = parseOptionalFieldWith pj
-
-#if !MIN_VERSION_base(4,16,0)
-instance {-# INCOHERENT #-} LookupField (Semigroup.Option a) where
-    lookupField pj tName rec obj key =
-        fmap Semigroup.Option
-             (lookupField (fmap Semigroup.getOption . pj) tName rec obj key)
-#endif
-
-lookupFieldWith :: (Value -> Parser a) -> String -> String
-                -> Object -> Key -> Parser a
-lookupFieldWith pj tName rec obj key =
+lookupFieldNoOmit :: (Value -> Parser a) -> String -> String -> Object -> Key -> Parser a
+lookupFieldNoOmit pj tName rec obj key =
     case KM.lookup key obj of
       Nothing -> unknownFieldFail tName rec (Key.toString key)
       Just v  -> pj v <?> Key key
@@ -1257,11 +1212,7 @@ deriveJSONClass consFuns jc opts name = do
   case info of
     DatatypeInfo { datatypeContext   = ctxt
                  , datatypeName      = parentName
-#if MIN_VERSION_th_abstraction(0,3,0)
                  , datatypeInstTypes = instTys
-#else
-                 , datatypeVars      = instTys
-#endif
                  , datatypeVariant   = variant
                  , datatypeCons      = cons
                  } -> do
@@ -1293,11 +1244,7 @@ mkFunCommon consFun jc opts name = do
   case info of
     DatatypeInfo { datatypeContext   = ctxt
                  , datatypeName      = parentName
-#if MIN_VERSION_th_abstraction(0,3,0)
                  , datatypeInstTypes = instTys
-#else
-                 , datatypeVars      = instTys
-#endif
                  , datatypeVariant   = variant
                  , datatypeCons      = cons
                  } -> do
@@ -1307,20 +1254,26 @@ mkFunCommon consFun jc opts name = do
       !_ <- buildTypeInstance parentName jc ctxt instTys variant
       consFun jc parentName opts instTys cons
 
+data FunArg = Omit | Single | Plural deriving (Eq)
+
 dispatchFunByType :: JSONClass
                   -> JSONFun
                   -> Name
                   -> TyVarMap
-                  -> Bool -- True if we are using the function argument that works
-                          -- on lists (e.g., [a] -> Value). False is we are using
-                          -- the function argument that works on single values
-                          -- (e.g., a -> Value).
+                  -> FunArg -- Plural if we are using the function argument that works
+                            -- on lists (e.g., [a] -> Value). Single is we are using
+                            -- the function argument that works on single values
+                            -- (e.g., a -> Value). Omit if we use it to check omission
+                            -- (e.g. a -> Bool)
                   -> Type
                   -> Q Exp
 dispatchFunByType _ jf _ tvMap list (VarT tyName) =
     varE $ case M.lookup tyName tvMap of
-                Just (tfjExp, tfjlExp) -> if list then tfjlExp else tfjExp
-                Nothing                -> jsonFunValOrListName list jf Arity0
+                Just (tfjoExp, tfjExp, tfjlExp) -> case list of
+                    Omit -> tfjoExp
+                    Single -> tfjExp 
+                    Plural -> tfjlExp
+                Nothing                   -> jsonFunValOrListName list jf Arity0
 dispatchFunByType jc jf conName tvMap list (SigT ty _) =
     dispatchFunByType jc jf conName tvMap list ty
 dispatchFunByType jc jf conName tvMap list (ForallT _ _ ty) =
@@ -1339,24 +1292,29 @@ dispatchFunByType jc jf conName tvMap list ty = do
         tyVarNames :: [Name]
         tyVarNames = M.keys tvMap
 
+        args :: [Q Exp]
+        args
+            | list == Omit = map     (dispatchFunByType jc jf conName tvMap  Omit)                        rhsArgs
+            | otherwise    = zipWith (dispatchFunByType jc jf conName tvMap) (cycle [Omit,Single,Plural]) (triple rhsArgs)
+
     itf <- isInTypeFamilyApp tyVarNames tyCon tyArgs
     if any (`mentionsName` tyVarNames) lhsArgs || itf
        then outOfPlaceTyVarError jc conName
        else if any (`mentionsName` tyVarNames) rhsArgs
-            then appsE $ varE (jsonFunValOrListName list jf $ toEnum numLastArgs)
-                         : zipWith (dispatchFunByType jc jf conName tvMap)
-                                   (cycle [False,True])
-                                   (interleave rhsArgs rhsArgs)
+            then appsE $ varE (jsonFunValOrListName list jf $ toEnum numLastArgs) : args
             else varE $ jsonFunValOrListName list jf Arity0
 
-dispatchToJSON
-  :: ToJSONFun -> JSONClass -> Name -> TyVarMap -> Type -> Q Exp
-dispatchToJSON target jc n tvMap =
-    dispatchFunByType jc (targetToJSONFun target) n tvMap False
+dispatchToJSON :: ToJSONFun -> JSONClass -> Name -> TyVarMap -> Type -> Q Exp
+dispatchToJSON target jc n tvMap = dispatchFunByType jc (targetToJSONFun target) n tvMap Single
 
-dispatchParseJSON
-  :: JSONClass -> Name -> TyVarMap -> Type -> Q Exp
-dispatchParseJSON  jc n tvMap = dispatchFunByType jc ParseJSON  n tvMap False
+dispatchOmitField :: JSONClass -> Name -> TyVarMap -> Type -> Q Exp
+dispatchOmitField jc n tvMap = dispatchFunByType jc ToJSON n tvMap Omit
+
+dispatchParseJSON :: JSONClass -> Name -> TyVarMap -> Type -> Q Exp
+dispatchParseJSON  jc n tvMap = dispatchFunByType jc ParseJSON  n tvMap Single
+
+dispatchOmittedField :: JSONClass -> Name -> TyVarMap -> Type -> Q Exp
+dispatchOmittedField jc n tvMap = dispatchFunByType jc ParseJSON n tvMap Omit
 
 --------------------------------------------------------------------------------
 -- Utility functions
@@ -1456,9 +1414,7 @@ buildTypeInstance tyConName jc dataCxt varTysOrig variant = do
                          Newtype         -> False
                          DataInstance    -> True
                          NewtypeInstance -> True
-#if MIN_VERSION_th_abstraction(0,5,0)
-                         Datatype.TypeData -> False
-#endif
+                         Language.Haskell.TH.Datatype.TypeData -> False
 
         remainingTysOrigSubst' :: [Type]
         -- See Note [Kind signatures in derived instances] for an explanation
@@ -1629,13 +1585,14 @@ Both.
 -- A mapping of type variable Names to their encoding/decoding function Names.
 -- For example, in a ToJSON2 declaration, a TyVarMap might look like
 --
--- { a ~> (tj1, tjl1)
--- , b ~> (tj2, tjl2) }
+-- { a ~> (o1, tj1, tjl1)
+-- , b ~> (o2, tj2, tjl2) }
 --
--- where a and b are the last two type variables of the datatype, tj1 and tjl1 are
--- the function arguments of types (a -> Value) and ([a] -> Value), and tj2 and tjl2
--- are the function arguments of types (b -> Value) and ([b] -> Value).
-type TyVarMap = Map Name (Name, Name)
+-- where a and b are the last two type variables of the datatype,
+-- o1 and o2 are function argument of types (a -> Bool),
+-- tj1 and tjl1 are the function arguments of types (a -> Value)
+-- and ([a] -> Value), and tj2 and tjl2 are the function arguments of types (b -> Value) and ([b] -> Value).
+type TyVarMap = Map Name (Name, Name, Name)
 
 -- | Returns True if a Type has kind *.
 hasKindStar :: Type -> Bool
@@ -1659,7 +1616,7 @@ newNameList prefix len = mapM newName [prefix ++ show n | n <- [1..len]]
 hasKindVarChain :: Int -> Type -> Maybe [Name]
 hasKindVarChain kindArrows t =
   let uk = uncurryKind (tyKind t)
-  in if (NE.length uk - 1 == kindArrows) && F.all isStarOrVar uk
+  in if (NE.length uk - 1 == kindArrows) && all isStarOrVar uk
         then Just (concatMap freeVariables uk)
         else Nothing
 
@@ -1680,9 +1637,11 @@ varTToNameMaybe _          = Nothing
 varTToName :: Type -> Name
 varTToName = fromMaybe (error "Not a type variable!") . varTToNameMaybe
 
-interleave :: [a] -> [a] -> [a]
-interleave (a1:a1s) (a2:a2s) = a1:a2:interleave a1s a2s
-interleave _        _        = []
+flatten3 :: [(a,a,a)] -> [a]
+flatten3 = foldr (\(a,b,c) xs -> a:b:c:xs) []
+
+triple :: [a] -> [a]
+triple = foldr (\x xs -> x:x:x:xs) []
 
 -- | Fully applies a type constructor to its type variables.
 applyTyCon :: Name -> [Type] -> Type
@@ -1714,17 +1673,10 @@ isInTypeFamilyApp names tyFun tyArgs =
     go tcName = do
       info <- reify tcName
       case info of
-#if MIN_VERSION_template_haskell(2,11,0)
         FamilyI (OpenTypeFamilyD (TypeFamilyHead _ bndrs _ _)) _
           -> withinFirstArgs bndrs
         FamilyI (ClosedTypeFamilyD (TypeFamilyHead _ bndrs _ _) _) _
           -> withinFirstArgs bndrs
-#else
-        FamilyI (FamilyD TypeFam _ bndrs _) _
-          -> withinFirstArgs bndrs
-        FamilyI (ClosedTypeFamilyD _ bndrs _ _) _
-          -> withinFirstArgs bndrs
-#endif
         _ -> return False
       where
         withinFirstArgs :: [a] -> Q Bool
@@ -1763,12 +1715,7 @@ mentionsName = go
 
 -- | Does an instance predicate mention any of the Names in the list?
 predMentionsName :: Pred -> [Name] -> Bool
-#if MIN_VERSION_template_haskell(2,10,0)
 predMentionsName = mentionsName
-#else
-predMentionsName (ClassP n tys) names = n `elem` names || any (`mentionsName` names) tys
-predMentionsName (EqualP t1 t2) names = mentionsName t1 names || mentionsName t2 names
-#endif
 
 -- | Split an applied type into its individual components. For example, this:
 --
@@ -1846,12 +1793,7 @@ valueConName (Bool   _) = "Boolean"
 valueConName Null       = "Null"
 
 applyCon :: Name -> Name -> Pred
-applyCon con t =
-#if MIN_VERSION_template_haskell(2,10,0)
-          AppT (ConT con) (VarT t)
-#else
-          ClassP con [VarT t]
-#endif
+applyCon con t = AppT (ConT con) (VarT t)
 
 -- | Checks to see if the last types in a data family instance can be safely eta-
 -- reduced (i.e., dropped), given the other types. This checks for three conditions:
@@ -1990,6 +1932,17 @@ jsonClassName (JSONClass From Arity0) = ''FromJSON
 jsonClassName (JSONClass From Arity1) = ''FromJSON1
 jsonClassName (JSONClass From Arity2) = ''FromJSON2
 
+jsonFunOmitName :: JSONFun -> Arity -> Name
+jsonFunOmitName ToJSON     Arity0 = 'omitField
+jsonFunOmitName ToJSON     Arity1 = 'liftOmitField
+jsonFunOmitName ToJSON     Arity2 = 'liftOmitField2
+jsonFunOmitName ToEncoding Arity0 = 'omitField
+jsonFunOmitName ToEncoding Arity1 = 'liftOmitField
+jsonFunOmitName ToEncoding Arity2 = 'liftOmitField2
+jsonFunOmitName ParseJSON  Arity0 = 'omittedField
+jsonFunOmitName ParseJSON  Arity1 = 'liftOmittedField
+jsonFunOmitName ParseJSON  Arity2 = 'liftOmittedField2
+
 jsonFunValName :: JSONFun -> Arity -> Name
 jsonFunValName ToJSON     Arity0 = 'toJSON
 jsonFunValName ToJSON     Arity1 = 'liftToJSON
@@ -2012,10 +1965,11 @@ jsonFunListName ParseJSON  Arity0 = 'parseJSONList
 jsonFunListName ParseJSON  Arity1 = 'liftParseJSONList
 jsonFunListName ParseJSON  Arity2 = 'liftParseJSONList2
 
-jsonFunValOrListName :: Bool -- e.g., toJSONList if True, toJSON if False
+jsonFunValOrListName :: FunArg -- e.g., toJSONList if True, toJSON if False
                      -> JSONFun -> Arity -> Name
-jsonFunValOrListName False = jsonFunValName
-jsonFunValOrListName True  = jsonFunListName
+jsonFunValOrListName Omit   = jsonFunOmitName
+jsonFunValOrListName Single = jsonFunValName
+jsonFunValOrListName Plural = jsonFunListName
 
 arityInt :: JSONClass -> Int
 arityInt = fromEnum . arity
